#include <iostream>

#include <Eigen/Dense>

using Matrix = Eigen::MatrixXd;
using Vector = Eigen::VectorXd;

//! \brief Use symmetric Gauss-Seidel iterations to approximate the solution of the system Ax = b
//! \param[in] A system matrix to be decomposed, must be invertible (L + D + U)
//! \param[in] b r.h.s. vector
//! \param[in,out] x initial guess as input and last value of iteration as output
//! \param[in] rtol relative tolerance for termination criteria
void GSIt(const Matrix & A, const Vector & b, Vector & x, double rtol) {
    
    // Extract strictly triangular part
    auto U = Eigen::TriangularView<Matrix, Eigen::StrictlyUpper>(A);
    auto L = Eigen::TriangularView<Matrix, Eigen::StrictlyLower>(A);
    
    // Extract non-strictly triangular parts (we only require L+D and U+D)
    auto UpD = Eigen::TriangularView<Matrix, Eigen::Upper>(A);
    auto LpD = Eigen::TriangularView<Matrix, Eigen::Lower>(A);
    
    // Temporary storage
    Vector temp(x.size());
    Vector* xold = &x;
    Vector* xnew = &temp;
    
    // Counter for iterations
    unsigned int k = 0;
    // Error of this and previous iteration
    double err;
//     double errold = 0.;
    do {
        // Apply iteration scheme
        *xnew = UpD.solve(b) - UpD.solve(L*LpD.solve(b - U**xold));
        //Compute error
        err = (*xold - *xnew).norm();
        // Output error and rate for problem 1e
//         std::cout << k++ << "\t& " << (A*(*xnew) - b).norm() << "\t& " << (*xold - *xnew).norm() / errold << "\t\\\\" << std::endl;
        
        // Proceed to next step
        std::swap(xold, xnew);
//         errold = err;
    } while( err > rtol*(*xnew).norm() ); // Termination criteria when tolerance reached
    
    x = *xnew; // Output was last value of iterations
    
    return;
}

int main(int, char**) {
    // Build matrix and r.h.s.
    unsigned int n = 9;
    
    Matrix A = Matrix::Zero(n,n);
    for(unsigned int i = 0; i < n; ++i) {
        if(i > 0) A(i,i-1) = 2;
        A(i,i) = 3;
        if(i < n-1) A(i,i+1) = 1;
    }
    Vector b = Vector::Constant(n,1);
    
    //// PROBLEM 1d
    std::cout << "*** PROBLEM 1d:" << std::endl;
    
    Vector x = b;
    GSIt(A, b, x, 10e-8);
    
    double residual = (A*x - b).norm();
    
    std::cout << "Residual = " << residual << std::endl;
}

