#include <iostream>

#include <vector>

#include <Eigen/Sparse>
#include <Eigen/SparseLU>

#include "timer.h"

using Triplet = Eigen::Triplet<double>;
using Triplets = std::vector<Triplet>;

using Vector = Eigen::VectorXd;
using Matrix = Eigen::SparseMatrix<double>;

//! \brief Efficiently construct the sparse matrix A given c, i_0 and j_0
//! \param[in] c contains entries c_i for matrix A
//! \param[in] i0 row index i_0
//! \param[in] j0 column index j_0
//! \return Sparse matrix A
Matrix buildA(const Vector & c, unsigned int i0, unsigned int j0) {
    assert(i0 > j0);
    
    unsigned int n = c.size() + 1;
    
    Triplets triplets;
    
    unsigned int ntriplets = 2*n;
    
    // TODO: problem 2a, construct the matrix A given c, i0 and j0
    
    // Reserve space
    triplets.reserve(ntriplets);
    
    // Build triplets vector
    for(unsigned int i = 0; i < n; ++i) {
        triplets.push_back(Triplet(i,i,1));
        if(i < n-1) triplets.push_back(Triplet(i,i+1,c[i]));
    }
    triplets.push_back(Triplet(i0,j0,1));
    
    // Construct sparse matrix from its triplets
    Matrix A(n,n);
    A.setFromTriplets(triplets.begin(), triplets.end());
    
    return A;
}

//! \brief Solve system Ax = b with optimal complexity O(n)
//! \param[in] c Entries for matrix A
//! \param[in] b r.h.s. vector
//! \param[in] i0 index
//! \param[in] j0 index
//! \return Solution x, s.t. Ax = b
template <unsigned int variant = 1>
Vector solveLSE(const Vector & c, const Vector & b, unsigned int i0, unsigned int j0) {
    assert(c.size() == b.size()-1 && "Size mismatch!");
    assert(i0 > j0);
    
    //// PROBLEM 2b
    
    // Allocate solution vector
    Vector ret(b.size());
    
    // VARIANT 1: simple Gauss elimination
    if(variant == 1) {
        ret(b.size()-1) = b(b.size()-1);
        
        // Loop all rows of matrix backwards
        for(int i = b.size()-2; i >= 0; --i) {
            // If row i0 is reached eliminate pivot
            if(i == (int) i0) {
                double piv = 1;
                double rhs = b(i);
                int sign = -1;
                for(unsigned int j = j0; j < i0; ++j) {
                    rhs -= piv * b(j);
                    piv *= sign*c(j);
                    sign *= -1;
                }
                ret(i) = (rhs - ret(i+1)*c(i)) / (1.-piv);
            } else { // Otherwise just do backwards substitution
                ret(i) = b(i) - ret(i+1)*c(i);
            }
        }
    // VARIANT 2: Sherman-Morrison-Woodbury formula
    } else if(variant == 2) {
        unsigned int n = b.size();

        Triplets triplets;

        unsigned int ntriplets = 2*n;

        // Reserve space
        triplets.reserve(ntriplets);

        // Build triplets vector
        for(unsigned int i = 0; i < n; ++i) {
            triplets.push_back(Triplet(i,i,1));
            if(i < n-1) triplets.push_back(Triplet(i,i+1,c[i]));
        }
        
        // Construct sparse matrix from its triplets
        Matrix A(n,n);
        A.setFromTriplets(triplets.begin(), triplets.end());

        Vector u = Vector::Zero(n);
        Eigen::MatrixXd v = Eigen::MatrixXd::Zero(1,n);
        u( i0 ) = 1;
        v( 0, j0 ) = 1;
        
        Vector Ainv_b = A.triangularView<Eigen::Upper>().solve( b );
        ret = Ainv_b - A.triangularView<Eigen::Upper>().solve( u * (v * Ainv_b)(0) ) / (1. + (v * A.triangularView<Eigen::Upper>().solve( u ))(0) );
    }
    
    return ret;
}

int main(int, char**) {
    
    Vector slv, ver1, ver2;
    
    Timer t_slv, t_ver1, t_ver2;
    for(unsigned int n = 2; n < 1024; n = n << 1) {
        
        unsigned int i0 = std::rand() % n, j0 = std::rand() % i0;
            
        Vector b = Vector::Random(n); // Random vector for b
        Vector c = Vector::Random(n-1); // Random vector for c
        
        t_slv.start();
        Matrix A = buildA(c, i0, j0);
        
        // Solve sparse system using sparse LU and our own routine
        A.makeCompressed();
        Eigen::SparseLU<Matrix> splu;
        splu.analyzePattern(A); 
        splu.factorize(A);
    
        slv = splu.solve(b); 
        t_slv.stop();
        
        t_ver1.start();
        ver1 = solveLSE<1>(c,b,i0,j0);
        t_ver1.stop();
        
        
        t_ver2.start();
        ver2 = solveLSE<2>(c,b,i0,j0);
        t_ver2.stop();
        
        std::cout << n << "\t" << (slv-ver1).norm() << "\t" << (slv-ver2).norm() << "\t"<< t_slv.duration() << "\t" << t_ver1.duration() << "\t" << t_ver2.duration() << std::endl;
    }
}
