#include <iostream>

#include <vector>

#include <Eigen/Dense>

using Vector = Eigen::VectorXd;
using Matrix = Eigen::MatrixXd;

//! \brief Evolves the vector y0 using the evolution operator \tilde{\Psi} with step-size y0
//! \tparam Operator type for evolution operator (e.g. lambda function type)
//! \param[in] Psi original evolution operator, must have operator(double, const Vector&)
//! \param[in] p parameter p for construction of Psi tilde
//! \param[in] h step-size
//! \param[in] y0 previous step
//! \return Evolved step \f$ \tilde{\Psi}^h y0 \f$
template <class Operator>
Vector psitilde(const Operator& Psi, unsigned int p, double h, const Vector & y0) {
    return ( Psi(h,y0) - (2<<(p))*Psi(h/2., Psi(h/2.,y0)) ) / (1. - (2<<(p)));
}

//! \brief Evolves the vector y0 using the evolution operator from time 0 to T using equidistant steps
//! \tparam Operator type for evolution operator (e.g. lambda function type)
//! \param[in] Psi original evolution operator, must have operator(double, const Vector&)
//! \param[in] T final time
//! \param[in] y0 initial data
//! \param[in] N number of steps
//! \return Vector of all steps y_0, y_1, ...
template <class Operator>
std::vector<Vector> odeintequi(const Operator& Psi, double T, const Vector &y0, int N) {
    double h = T / N;
    double t = 0.;
    std::vector<Vector> Y;
    Y.reserve(N+1);
    Y.push_back(y0);
    Vector y = y0;
    
    while( t < T ) {
        y = Psi(h,Y.back());
        
        Y.push_back(y);
        t += std::min(T-t,h);
    }
    
    return Y;
}

//! \brief Evolves the vector y0 using the evolution operator from time 0 to T using adaptive error control
//! \tparam Operator type for evolution operator (e.g. lambda function type)
//! \param[in] Psi original evolution operator, must have operator(double, const Vector&)
//! \param[in] T final time
//! \param[in] y0 initial data
//! \param[in] h0 initial step size
//! \param[in] p parameter p for construction of Psi tilde
//! \param[in] reltol relative tolerance for error control
//! \param[in] abstol absolute tolerance for error control
//! \param[in] p parameter p for construction of Psi tilde
//! \param[in] hmin minimal step size
//! \return Vector of all steps y_0, y_1, ...
template <class Operator>
std::vector<Vector> odeintssctrl(const Operator& Psi, double T, const Vector &y0, double h0,
                                 unsigned int p, double reltol, double abstol, double hmin) {
    // Time tracker
    double t = 0.;
    // Step size tracker
    double h = h0;
    // Vector of snapshots
    std::vector<Vector> Y;
    // Save initial data
    Y.push_back(y0);
    // Start with y0 as value
    Vector y = y0;
    
    // Kill if final time reached or min step size reached
    while( t < T && h > hmin ) {
        // Compute with two methods
        Vector y_high = psitilde(Psi, p, h, y);
        Vector y_low = Psi(h,y);
        // Estimate error
        double err_est = (y_high - y_low).norm();
        
        // Compute tolerance (min of reltol and abstol)
        double tol = std::max(reltol*y_high.norm(),abstol);
        
        // Accept step if true (error smaller than tol)
        if( err_est < tol ) {
            t += h;
            
            y = y_high;
            Y.push_back(y_high);
            
        }
        // Predict step-size
        h *= std::max(.5, std::min(2., std::pow(tol/err_est, 1./(p+1.))));
        // Force time if needed
        h = std::min(T-t, h);
    }
    if( t < T ) std::cerr << "Failed to reach final time!" << std::endl;
    
    return Y;
}

int main(int, char**) {
    
    auto f = [] (const Vector &y) -> Vector { return Vector::Ones(1) + y*y; };
    Vector y0 = Vector::Zero(1);
    double T = 1.;
    auto y_ex = [] (double t) -> Vector { Vector y(1); y << tan(t); return y; };
    
    unsigned p = 1;
    
    auto Psi = [&f] (double h, const Vector & y0) -> Vector { return y0 + h*f(y0); };
    auto PsiTilde = [&Psi, &f, &p] (double h, const Vector & y0) -> Vector {  return psitilde(Psi, p, h, y0); };
    
    //// PROBLEM 5d
    std::cout << "*** PROBLEM 5d:" << std::endl;
    
    std::cout << "Error table for equidistant steps:" << std::endl;
    std::cout << "N" << "\t" << "Error" << std::endl;
    for(int N = 4; N < 4096; N=N<<1 ) {
        std::vector<Vector> Y = odeintequi(PsiTilde, T, y0, N);
        double err = (Y.back() - y_ex(1)).norm();
        std::cout << N << "\t" << err << std::endl;
    }
    
    //// PROBLEM 5e
    std::cout << "*** PROBLEM 5e:" << std::endl;
    
    double h0 = 1./100.;
    std::vector<Vector> Y;
    double err = 0;
    Y = odeintssctrl(Psi, T, y0, h0, p, 10e-8, 10e-8, 10e-6);
    err = (Y.back() - y_ex(1)).norm();
    std::cout << "Adaptive error control results:" << std::endl;
    std::cout << "Error" << "\t" << "No. Steps" << "\t" << "y(1)" << "\t" << "y(ex)" << std::endl;
    std::cout << err << "\t" << Y.size() << "\t" << Y.back() << "\t" << y_ex(1) << std::endl;
}
